import os
import sys
sys.path.insert(1, './secure_code')
from random import gauss
from Crypto.Hash import SHAKE256
from ntrugen import ntru_solve, gs_norm
from ntt import ntt, intt, div_zq, mul_zq, add_zq
from fft import fft, ifft, sub, neg, add_fft, mul_fft
from common import q
from math import sqrt

def prvkey_gen(n = 128):
    while True:
        sigma = 1.17 * sqrt(q / (2 * n))
        f = [int(gauss(0, sigma)) for _ in range(n)]
        g = [int(gauss(0, sigma)) for _ in range(n)]
        if gs_norm(f, g, q) > (1.17 ** 2) * q:
            continue
        f_ntt = ntt(f)
        if any((elem == 0) for elem in f_ntt):
            continue
        try:
            F, G = ntru_solve(f, g)
            F = [int(coef) for coef in F]
            G = [int(coef) for coef in G]
            return f, g, F, G
        except ValueError:
            continue

def pubkey_gen(sk):
    f, g, _, _ = sk
    return div_zq(g, f)

def hash_to_point(salt, message, n = 128):
    k = (1 << 16) // q
    shake = SHAKE256.new()
    shake.update(salt)
    shake.update(message)
    hashed = [0 for i in range(n)]
    i = 0
    j = 0
    while i < n:
        twobytes = shake.read(2)
        elt = (twobytes[0] << 8) + twobytes[1]
        if elt < k * q:
            hashed[i] = elt % q
            i += 1
        j += 1
    return hashed

# this function has been HIGHLY optimized to be super efficient
def sign(sk, message):
    f, g, F, G = sk
    B0 = [
        [g, neg(f)],
        [G, neg(F)],
    ]

    r = os.urandom(40)
    point = hash_to_point(r, message)
    n = len(point)

    B0_fft = [[fft(elt) for elt in row] for row in B0]
    [[a, b], [c, d]] = B0_fft

    point_fft = fft(point)
    t0_fft = [(point_fft[i] * d[i]) / q for i in range(n)]
    t1_fft = [(-point_fft[i] * b[i]) / q for i in range(n)]

    z0 = [round(elt) for elt in ifft(t0_fft)]
    z1 = [round(elt) for elt in ifft(t1_fft)]

    z_fft = [fft(z0), fft(z1)]
   
    v0_fft = add_fft(mul_fft(z_fft[0], a), mul_fft(z_fft[1], c))
    v1_fft = add_fft(mul_fft(z_fft[0], b), mul_fft(z_fft[1], d))
    v0 = [int(round(elt)) for elt in ifft(v0_fft)]
    v1 = [int(round(elt)) for elt in ifft(v1_fft)]

    s = [sub(point, v0), neg(v1)]
    return r, s

def verify(h, message, signature):
    r, s = signature
    if len(r) != 40:
        return False
    s0, s1 = s
    n = len(s0)
    signature_bound = 2 * n * q

    hashed = hash_to_point(r, message)
    point = add_zq(s0, mul_zq(s1, h))

    if point != hashed:
        return False

    norm_sign  = sum(coef ** 2 for coef in s0)
    norm_sign += sum(coef ** 2 for coef in s1)
    return norm_sign <= signature_bound

def test():
    sk = prvkey_gen()
    pk = pubkey_gen(sk)

    for _ in range(256):
        message = os.urandom(32)
        sig = sign(sk, message)
        assert verify(pk, message, sig)

if __name__ == "__main__":
    test()
